// Mark Kamichoff
// kamicm@rpi.edu
// Final Project

#include "defines.h"

// Some globals
int win_width, win_height;
GLUquadricObj* gluObj;

// Inits the external object structure
external_objects extobj[NUMOBJS];

// Normal calculations
float crossresult[3];

// Lights (-1 is off, 1 is on)
short lite0 = 1;
short lite1 = -1;
short lite2 = -1;
short lite3 = -1;

// misc -1 is keyb, 1 is mouse
GLint nNumPoints = 4;

// View Stuff (bool_shipview {-1=follow,1=lookfromcenter}
short bool_shipview = -1;

short bool_angy=0;
short bool_movez=0;
short bool_movey=0;
short drawstyle = 0;

GLfloat ratio;
short orbit_xang_bool=0;
short orbit_yang_bool=0;
short orbit_zang_bool=0;
short orbit_zoom_bool=0;
GLfloat orbit_xang = 75;
GLfloat orbit_yang = 25;
GLfloat orbit_zoom = 2;

GLfloat gradual_x = -2.0;
GLfloat gradual_y = 0;
GLfloat gradual_z = -1.0;
GLfloat gradual_center_x = 0;
GLfloat gradual_center_y = -1.0;
GLfloat gradual_center_z = 0;

// Selection variables
// Maximum of 32 ships, and { xpos,ypos,zpos,xang,yang,zang,special,visible,size }
// A value of -1 for selected_ship means nothing is selected and the camera
// view should be orbiting and walking
GLfloat shipdb[32][9];
short selected_ship=-1;

// Read in the selected models that will be used
void readextobjs(void)
{
//	read_raw_obj("models/freaky.raw", 0);
//	read_raw_obj("models/sphere.raw", 1);
//	read_raw_obj("models/square.raw", 2);
//	read_raw_obj("models/cylinder.raw", 3);
//	read_raw_obj("models/coolio.raw", 4);
//	read_raw_obj("models/coolio2.raw", 5);
//	read_raw_obj("models/toobig.raw", 6);
//	read_raw_obj("models/starbase.raw", 7);
//	read_raw_obj("models/blah.raw", 8);
	read_raw_obj("models/ywing.raw", 0);
	read_raw_obj("models/tiebomber.raw", 1);
//	read_raw_obj("models/deathstar.raw", 2);
	read_raw_obj("models/sd.raw", 2);
}

// Calculate the cross product of triangles from RAW vertices
void cross_prod(int id, int i)
{
	float v1[3];
	float v2[3];
	float length;
	
	v1[0] = extobj[id].vx[i] - extobj[id].vx[i+1];
	v1[1] = extobj[id].vy[i] - extobj[id].vy[i+1];
	v1[2] = extobj[id].vz[i] - extobj[id].vz[i+1];
	v2[0] = extobj[id].vx[i+1] - extobj[id].vx[i+2];
	v2[1] = extobj[id].vy[i+1] - extobj[id].vy[i+2];
	v2[2] = extobj[id].vz[i+1] - extobj[id].vz[i+2];
	crossresult[0] = v1[1]*v2[2] - v2[1]*v1[2];
	crossresult[1] = v2[0]*v1[2] - v1[0]*v2[2];
	crossresult[2] = v1[0]*v2[1] - v2[0]*v1[1];

	length = sqrt(crossresult[0]*crossresult[0]
			+ crossresult[1]*crossresult[1]
			+ crossresult[2]*crossresult[2]);

	if(length == 0.0)
		length = 1.0;

	crossresult[0] /= length;
	crossresult[1] /= length;
	crossresult[2] /= length;
}

// Draw an object by ID (Each RAW file has an ID)
void drawextobj(int id, GLfloat factor)
{
	int i;
	
	for(i=0; i<extobj[id].numvertices; i+=3)
	{
		glBegin(GL_TRIANGLES);

			cross_prod(id,i);
			glNormal3f(crossresult[0],crossresult[1],crossresult[2]);

			glVertex3f(extobj[id].vx[i]*factor,
					extobj[id].vy[i]*factor,
					extobj[id].vz[i]*factor);

			glVertex3f(extobj[id].vx[i+1]*factor,
					extobj[id].vy[i+1]*factor,
					extobj[id].vz[i+1]*factor);

			glVertex3f(extobj[id].vx[i+2]*factor,
					extobj[id].vy[i+2]*factor,
					extobj[id].vz[i+2]*factor);

		glEnd();
	}
}

void updatevalues(void)
{
	// For selected_ship == -1
	if(selected_ship == -1)
	{
		if(orbit_xang_bool == 1)
			orbit_xang += 1;
		else if(orbit_xang_bool == -1)
			orbit_xang -= 1;
	
		if(orbit_yang_bool == 1)
			orbit_yang += 1;
		else if(orbit_yang_bool == -1)
			orbit_yang -= 1;
	
		if(orbit_zoom_bool == 1)
			orbit_zoom += 0.01;
		else if(orbit_zoom_bool == -1)
			orbit_zoom -= 0.01;
	}
	else
	{
		if(bool_angy == 1)
			shipdb[selected_ship][4] += 1;
		else if(bool_angy == -1)
			shipdb[selected_ship][4] -= 1;

		if(bool_movez == 1)
		{
			shipdb[selected_ship][0] +=
				cos(PI*((int)(shipdb[selected_ship][4])%360)/180) * 0.01;
			shipdb[selected_ship][2] +=
				sin(PI*((int)(shipdb[selected_ship][4])%360)/180) * 0.01;
		}
		else if(bool_movez == -1)
		{
			shipdb[selected_ship][0] -=
				cos(PI*((int)(shipdb[selected_ship][4])%360)/180) * 0.01;
			shipdb[selected_ship][2] -=
				sin(PI*((int)(shipdb[selected_ship][4])%360)/180) * 0.01;
		}

		if(bool_movey == 1)
			shipdb[selected_ship][1] += 0.01;
		else if(bool_movey == -1)
			shipdb[selected_ship][1] -= 0.01;
	}

	// Update lighting
	if(lite0 == 1) glEnable(GL_LIGHT0);
	else glDisable(GL_LIGHT0);
	if(lite1 == 1) glEnable(GL_LIGHT1);
	else glDisable(GL_LIGHT1);
	if(lite2 == 1) glEnable(GL_LIGHT2);
	else glDisable(GL_LIGHT2);
	if(lite3 == 1) glEnable(GL_LIGHT3);
	else glDisable(GL_LIGHT3);
}

// Main display function
void display(void)
{
	int i;
	GLfloat tempx, tempy, tempz;

	updatevalues();

	// Set miscellaneous GL stuffs to get a nice camera view
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	if(selected_ship != -1 && bool_shipview == -1)
		gluPerspective(60,ratio,.01,10);
	else
		gluPerspective(60,ratio,.01,10);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// If we are orbiting
	if(selected_ship == -1)
	{
		// Assign the eventual values to temp? first
                tempx=sin(PI*orbit_xang/180)*cos(PI*orbit_yang/180)*orbit_zoom;
                tempy=sin(PI*orbit_yang/180)*orbit_zoom;
                tempz=cos(PI*orbit_xang/180)*cos(PI*orbit_yang/180)*orbit_zoom;

		// This makes it smooth transition between 't''ing
		gradual_x += (tempx-gradual_x)/TRANSITION_FACTOR;
		gradual_y += (tempy-gradual_y)/TRANSITION_FACTOR;
		gradual_z += (tempz-gradual_z)/TRANSITION_FACTOR;

		// Makes the center point gradual too...biotch!
		gradual_center_x += (-gradual_center_x)/TRANSITION_FACTOR;
		gradual_center_y += (-gradual_center_y)/TRANSITION_FACTOR;
		gradual_center_z += (-gradual_center_z)/TRANSITION_FACTOR;

		gluLookAt(gradual_x, gradual_y, gradual_z,
				gradual_center_x, gradual_center_y, gradual_center_z,
				0.0, 1.0, 0.0);

	}
	// If we are looking at a ship
	else
	{
		// If we're following it from the back (default)
		if(bool_shipview == -1)
		{
			// Assign the eventual values to temp? first
			tempx = shipdb[selected_ship][0]-
				sin(PI*((int)(shipdb[selected_ship][4]+90)%360)/180)*1.2;
			tempy = shipdb[selected_ship][1]+0.6;
			tempz = shipdb[selected_ship][2]+
				cos(PI*((int)(shipdb[selected_ship][4]+90)%360)/180)*1.2;

			// Then make the smooth (it'll neveer get there)
			// transition between the initial value of gradual_?
			// to temp?
			gradual_x += (tempx-gradual_x)/TRANSITION_FACTOR;
			gradual_y += (tempy-gradual_y)/TRANSITION_FACTOR;
			gradual_z += (tempz-gradual_z)/TRANSITION_FACTOR;

			// Makes the center point gradual too...biotch!
			gradual_center_x += (shipdb[selected_ship][0]-gradual_center_x)/TRANSITION_FACTOR;
			gradual_center_y += (shipdb[selected_ship][1]-gradual_center_y)/TRANSITION_FACTOR;
			gradual_center_z += (shipdb[selected_ship][2]-gradual_center_z)/TRANSITION_FACTOR;

			gluLookAt(gradual_x, gradual_y, gradual_z,
					gradual_center_x, gradual_center_y, gradual_center_z,
					0.0, 1.0, 0.0);
		}
		// Gee, guess we're looking at it from 0,0,0, really nifty!
		else
		{
			// This makes it smooth transition between 't''ing
			gradual_x += (-gradual_x)/TRANSITION_FACTOR;
			gradual_y += (-gradual_y)/TRANSITION_FACTOR;
			gradual_z += (-gradual_z)/TRANSITION_FACTOR;

			// Makes the center point gradual too...biotch!
			gradual_center_x += (shipdb[selected_ship][0]-gradual_center_x)/TRANSITION_FACTOR;
			gradual_center_y += (shipdb[selected_ship][1]-gradual_center_y)/TRANSITION_FACTOR;
			gradual_center_z += (shipdb[selected_ship][2]-gradual_center_z)/TRANSITION_FACTOR;

			gluLookAt(gradual_x, gradual_y, gradual_z,
					gradual_center_x,
					gradual_center_y,
					gradual_center_z,
					0.0, 1.0, 0.0);
		}
	}

	// Draws the ships, manipulating the properties as we go
	for(i=0; i < NUMSHIPS; i++)
	{
		// First, check if we wanna draw it...
		if(shipdb[i][7])
		{
			glPushMatrix();
				// Translate a little bit from saved values
				glTranslatef(shipdb[i][0], shipdb[i][1], shipdb[i][2]);
	
				// Rotate the axes from saved values
				glRotatef(shipdb[i][3], 1.0, 0.0, 0.0);
				glRotatef(-shipdb[i][4]+90, 0.0, 1.0, 0.0);
				glRotatef(shipdb[i][5], 0.0, 0.0, 1.0);
		
				// Small Star Destroyer
				if(i == 0)
					stardestroyer(shipdb[i][8]);
				// TIE Fighter
				else if(i == 1)
					tiefighter(shipdb[i][8]);
				// TIE Fighter Right Side
				else if(i == 2)
					tiefighter(shipdb[i][8]);
				// Imperial Shuttle
				else if(i == 3)
					shuttle(shipdb[i][8], shipdb[i][6]);
				// Y-Wing Fighter
				else if(i == 4)
					drawextobj(0, shipdb[i][8]);
				// TIE Bomber
				else if(i == 5)
					drawextobj(1, shipdb[i][8]);
				// New Star Destroyer
				else if(i == 6)
					drawextobj(2, shipdb[i][8]);
			glPopMatrix();
		}
	}

	glutSwapBuffers();
}

// generic resize function, scales to W x H
void myReshape(GLsizei w, GLsizei h)
{
	win_width = w;
	win_height = h;

	glViewport(0,0,win_width,win_height);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1.0, 1.0, -1.0, 1.0, -10.0, 10.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glClearColor(0.0, 0.0, 0.0,0.0);
	glClear(GL_COLOR_BUFFER_BIT);

	display();
	glutSwapBuffers();
}

// Good Perspective this time
void reshape_perspective(GLsizei w, GLsizei h)
{
	// Set the global width and height
	win_width = w;
	win_height = h;

	// Sets the ratio of the resized window
	ratio = (GLfloat)w/(GLfloat)h;

	// Debug
	printf("Debug: Thanks for calling reshape_perspective()!\n");

	// Set the viewport to take advantage of the new window dimensions
	glViewport(0,0,(GLsizei)win_width,(GLsizei)win_height);
}

// generic initialization routine
void myinit(void)
{
	// Set up all the light properties and positions
	GLfloat light0_ambient[] = { 1.0, 1.0, 1.0, 1.0 };
	GLfloat light0_position[] = { 0.0, 2.0, 0.0, 0.0 };

	GLfloat light1_ambient[] = { 0.0, 0.0, 1.0, 1.0 };
	GLfloat light1_position[] = { 1.0, 0.0, 0.0, 0.0 };
	GLfloat light1_diffuse[] = { 0.0, 0.0, 1.0, 1.0 };

	GLfloat light2_ambient[] = { 1.0, 0.0, 0.0, 1.0 };
	GLfloat light2_position[] = { -1.0, 0.0, 0.0, 0.0 };
	GLfloat light2_diffuse[] = { 1.0, 0.0, 0.0, 1.0 };

	GLfloat light3_ambient[] = { 0.0, 1.0, 0.0, 1.0 };
	GLfloat light3_position[] = { 0.0, -1.0, 0.0, 0.0 };
	GLfloat light3_diffuse[] = { 0.0, 1.0, 0.0, 1.0 };

//	glEnable(GL_AUTO_NORMAL);

	// Initial conditions for the ships/objects
	init_ships();

	// Make a new quadric object which we will use
	gluObj = gluNewQuadric();

	glClearColor(0.0, 0.0, 0.0, 0.0);
	glShadeModel(GL_SMOOTH);
	
	// Apply the properties to the lights
	glLightfv(GL_LIGHT0, GL_AMBIENT, light0_ambient);
	glLightfv(GL_LIGHT0, GL_POSITION, light0_position);

	glLightfv(GL_LIGHT1, GL_AMBIENT, light1_ambient);
	glLightfv(GL_LIGHT1, GL_POSITION, light1_position);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
	glLightfv(GL_LIGHT2, GL_AMBIENT, light2_ambient);
	glLightfv(GL_LIGHT2, GL_POSITION, light2_position);
	glLightfv(GL_LIGHT2, GL_DIFFUSE, light2_diffuse);
	glLightfv(GL_LIGHT3, GL_AMBIENT, light3_ambient);
	glLightfv(GL_LIGHT3, GL_POSITION, light3_position);
	glLightfv(GL_LIGHT3, GL_DIFFUSE, light3_diffuse);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);

	// Play with Alpha Blending
//	glEnable(GL_BLEND);
//	glBlendFunc(GL_SRC_ALPHA, GL_ONE);

//	glutSwapBuffers();
}

void root_menu(int id)
{
	// Dummy Function
	if(id == 1)
		exit(0);
}

// If we type a key on the keyboad
void key(unsigned char k, int x, int y)
{
	// Keybindings
	if(k == 'q') exit(0);

	// Cycles through filled, wire, and point drawstyles
	if(k == 'p')
	{
		if(drawstyle < 2) drawstyle++;
		else drawstyle=0;

		if(!drawstyle)
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		else if(drawstyle == 1)
			glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		else if(drawstyle == 2)
			glPolygonMode(GL_FRONT_AND_BACK, GL_POINT);
	}

	// Tabs through the ships (TAB == 't' temporarily...)
	if(k == 't')
	{
		if(selected_ship == (NUMSHIPS-1))
			selected_ship = -1;
		else
			selected_ship++;

		if(selected_ship == -1)
			printf("Debug: selected_ship = (None)\n");
		else
			printf("Debug: selected_ship = %d\n", selected_ship);
	}

	// Toggles ship view between follow and centerlook
	// If not in orbit view
	if((selected_ship != -1) && (k == 'm'))
		bool_shipview *= -1;

	// To move the ship when in ship camera mode
	if(k == 'h') bool_angy = -1;
	if(k == 'l') bool_angy = 1;
	if(k == 'j') bool_movez = -1;
	if(k == 'k') bool_movez = 1;

	if(k == 'a') bool_movey = 1;
	if(k == 'z') bool_movey = -1;
}

// Keybindings for ones that turn off when unpressed
void keyup(unsigned char k, int x, int y)
{
	// Keybindings up

	// To move the ship when in ship camera mode
	if(k == 'h') bool_angy = -0;
	if(k == 'l') bool_angy = 0;
	if(k == 'j') bool_movez = -0;
	if(k == 'k') bool_movez = 0;

	if(k == 'a') bool_movey = 0;
	if(k == 'z') bool_movey = -0;
}

// special keys
void skey(unsigned char k, int x, int y)
{
	if(selected_ship == -1)
	{
		// Manage the whole rotation
		if(k == GLUT_KEY_LEFT) orbit_xang_bool = -1;
		if(k == GLUT_KEY_RIGHT) orbit_xang_bool = 1;
		if(k == GLUT_KEY_UP) orbit_yang_bool = 1;
		if(k == GLUT_KEY_DOWN) orbit_yang_bool = -1;
		if(k == GLUT_KEY_HOME) orbit_zoom_bool = -1;
		if(k == GLUT_KEY_END) orbit_zoom_bool = 1;
	}

	// bunch of Fx keys toggle lights
	if(k == GLUT_KEY_F1) lite0 *= -1;
	if(k == GLUT_KEY_F2) lite1 *= -1;
	if(k == GLUT_KEY_F3) lite2 *= -1;
	if(k == GLUT_KEY_F4) lite3 *= -1;
}

// Up versions of the above bindings
void skeyup(unsigned char k, int x, int y)
{
	if(selected_ship == -1)
	{
		// Manage the whole rotation
		if(k == GLUT_KEY_LEFT) orbit_xang_bool = 0;
		if(k == GLUT_KEY_RIGHT) orbit_xang_bool = 0;
		if(k == GLUT_KEY_UP) orbit_yang_bool = 0;
		if(k == GLUT_KEY_DOWN) orbit_yang_bool = -0;
		if(k == GLUT_KEY_HOME) orbit_zoom_bool = 0;
		if(k == GLUT_KEY_END) orbit_zoom_bool = 0;
	}
}

// Passive mouse...same as in last project
void mouse_passive(int x, int y)
{
	y = win_height - y;

//	tempx = x;
//	tempy = y;
}

// Main function
int main(int argc, char **argv)
{
	// Sets the global h/w
	win_width = INIT_W;
	win_height = INIT_H;

	ratio = (GLfloat)win_width/(GLfloat)win_height;

	// Reads in external object (RAW) files
	readextobjs();
		
	// GLUT stuff and create the menus
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowSize(INIT_W, INIT_H);
	glutCreateWindow("Final");

	glutDisplayFunc(display);
	glutIdleFunc(display);

	glutCreateMenu(root_menu);
	glutCreateMenu(root_menu);
	glutAddMenuEntry("Exits", 1);

	// right click for the menu
	glutAttachMenu(GLUT_RIGHT_BUTTON);

	myinit();

//	glutReshapeFunc(myReshape);
	glutReshapeFunc(reshape_perspective);

	glutIgnoreKeyRepeat(1);
	glutKeyboardFunc(key);
	glutKeyboardUpFunc(keyup);
	glutSpecialFunc((void *)skey);
	glutSpecialUpFunc((void *)skeyup);
	glutPassiveMotionFunc(mouse_passive);

	glutMainLoop();

	return 0;
}
